<?xml version="1.0" encoding="UTF-8"?>
<xsl:stylesheet xmlns:xsl="http://www.w3.org/1999/XSL/Transform" 
                xmlns:m="http://www.schema-library.org/mixed-content/restrict" 
                xmlns:xs="http://www.w3.org/2001/XMLSchema"
                exclude-result-prefixes="#all" 
                version="2.0">

    <!-- 
        This program deals with a complexType that restricts
        a complexType with mixed content. See my paper
        "Restricting complexTypes that have mixed content"
        
        Author: Roger L. Costello
    -->

    <xsl:key name="ids" match="*[@_id]" use="@_id"/>

    <xsl:function name="m:complexTypes-that-restrict-mixed-content" as="document-node()">
        <xsl:param name="schemas" as="document-node()"/>

        <xsl:variable name="updated-schemas">
            <schemas>
                <xsl:apply-templates select="$schemas/schemas/xs:schema" mode="complexTypes-that-restrict-mixed-content"/>
            </schemas>
        </xsl:variable>

        <xsl:sequence select="$updated-schemas"/>

    </xsl:function>

    <xsl:template match="xs:schema" mode="complexTypes-that-restrict-mixed-content">

        <xsl:copy>
            <xsl:copy-of select="@*"/>
            <xsl:apply-templates mode="complexTypes-that-restrict-mixed-content"/>
        </xsl:copy>

    </xsl:template>
    
    <!-- The base complexType C1 has mixed content. The derived complexType removes the text that can surround C1’s child elements.
         The derived complexType is equivalent to a complexType with child elements, so it is replaced by this. -->
    
    <xsl:template match="xs:complexType[xs:complexContent/xs:restriction[@base][key('ids', @idref)[@mixed eq 'true']]/not(xs:simpleType)]" mode="complexTypes-that-restrict-mixed-content">
        <xsl:copy>
            <xsl:copy-of select="@*"/>
            <xsl:apply-templates select="xs:complexContent/xs:restriction/node()" mode="complexTypes-that-restrict-mixed-content"/>
        </xsl:copy>
    </xsl:template>
   
    <!-- The base complexType C1 has mixed content and no attributes. The derived complexType removes the child elements of C1. 
         The derived complexType is equivalent to a simpleType, so it is replaced with a simpleType. -->

    <xsl:template match="xs:complexType[xs:simpleContent/xs:restriction[@base][key('ids', @idref)[@mixed eq 'true'][not(xs:attribute)][not(xs:attributeGroup)][not(xs:anyAttribute)]]/xs:simpleType]" mode="complexTypes-that-restrict-mixed-content">
        <xs:simpleType xmlns:xs="http://www.w3.org/2001/XMLSchema">
            <xsl:copy-of select="@*"/>
            <xsl:apply-templates select="xs:simpleContent/xs:restriction/xs:simpleType/node()" mode="complexTypes-that-restrict-mixed-content"/>
        </xs:simpleType>
    </xsl:template>

    <!-- The base complexType C2 has mixed content and has attributes. The derived complexType removes the child elements of C2. 
         The derived complexType is equivalent to a pair: (1) a simpleType and (2) a complexType with simpleContent that extends
         the simpleType with the attribute(s). -->

    <xsl:template match="xs:complexType[xs:simpleContent/xs:restriction[@base][key('ids', @idref)[@mixed eq 'true'][xs:attribute or xs:attributeGroup or xs:anyAttribute]]/xs:simpleType]" mode="complexTypes-that-restrict-mixed-content">
        <xsl:variable name="simpleType-name" select="generate-id(xs:simpleContent/xs:restriction/xs:simpleType)"/>

        <xs:simpleType name="{$simpleType-name}">
            <xsl:copy-of select="xs:simpleContent/xs:restriction/xs:simpleType/@*"/>
            <xsl:sequence select="xs:simpleContent/xs:restriction/xs:simpleType/child::node()" />
        </xs:simpleType>

        <xsl:copy>
            <xsl:copy-of select="@*"/>
            <xsl:for-each select="xs:simpleContent">
                <xsl:copy>
                    <xsl:copy-of select="@*"/>
                    <xs:extension base="{$simpleType-name}">
                        <xsl:sequence select="xs:restriction/(node() except xs:simpleType) " />
                    </xs:extension>
                </xsl:copy>
            </xsl:for-each>
        </xsl:copy>
    </xsl:template>


    <xsl:template match="node()" mode="complexTypes-that-restrict-mixed-content">
        
        <xsl:copy>
            <xsl:copy-of select="@*"/>
            <xsl:apply-templates mode="complexTypes-that-restrict-mixed-content"/>
        </xsl:copy>

    </xsl:template>

</xsl:stylesheet>
