<?xml version="1.0" encoding="UTF-8"?>
<xsl:stylesheet xmlns:xsl="http://www.w3.org/1999/XSL/Transform" 
                xmlns:s="http://www.schema-library.org/show/elements/type" 
                xmlns:xs="http://www.w3.org/2001/XMLSchema" 
                exclude-result-prefixes="#all" 
                version="2.0">

    <!-- 
        Consider an element E.
        
        E does not specify a type if it has 
        - no ref attribute, 
        - no type attribute,
        - no complexType child element, and 
        - no simpleType child element. 
        
        Here is a function that returns E's type.
        
        type (E) = if (E is local) then 
                       return xs:anyType
                   else   (: E is global :)
                       if (E has a substitutionGroup attribute to element S) then
                           return type (S)
                       else (: E is global with no substitutionGroup :)
                           return xs:anyType
        
        
        Author: Roger L. Costello
    -->
    
    <xsl:key name="ids" match="*[@_id]" use="@_id"/>

    <xsl:function name="s:add-type-to-elements-and-attributes-with-no-type-specified" as="document-node()">
        <xsl:param name="schemas" as="document-node()"/>

        <xsl:variable name="updated-schemas">
            <schemas>
                <xsl:apply-templates select="$schemas/schemas/xs:schema" mode="elements-and-attributes-with-no-type"/>
            </schemas>
        </xsl:variable>

        <xsl:sequence select="$updated-schemas"/>

    </xsl:function>


    <xsl:template match="xs:schema" mode="elements-and-attributes-with-no-type">

        <xsl:copy>
            <xsl:copy-of select="@*"/>
            <xsl:apply-templates mode="elements-and-attributes-with-no-type" />
        </xsl:copy>

    </xsl:template>
    
    
    <xsl:template match="xs:element[not(@ref)][not(@type)][not(xs:simpleType)][not(xs:complexType)][not(@substitutionGroup)]" mode="elements-and-attributes-with-no-type">
        <xsl:copy>
            <xsl:copy-of select="@*" />
            <xsl:sequence select="node()" />
            <xs:complexType mixed="true">
                <xs:sequence>
                    <xs:any minOccurs="0" maxOccurs="unbounded"/>
                </xs:sequence>
                <xs:anyAttribute/>
            </xs:complexType>
        </xsl:copy>
    </xsl:template>
    
    
    <xsl:template match="xs:element[not(@ref)][not(@type)][not(xs:simpleType)][not(xs:complexType)][@substitutionGroup]" mode="elements-and-attributes-with-no-type">
        <xsl:copy>
            <xsl:copy-of select="@*" />
            <xsl:sequence select="node()" />
            <xsl:variable name="r" select="s:type(key('ids', @substitutionGroupRef))" />
            <xsl:sequence select="s:type(key('ids', @substitutionGroupRef))" />
        </xsl:copy>
    </xsl:template>
    
    <xsl:function name="s:type">
        <xsl:param name="E" />
        
        <xsl:choose>
            <xsl:when test="$E/xs:simpleType">
                <xsl:sequence select="$E/xs:simpleType" />
            </xsl:when>
            <xsl:when test="$E/xs:complexType">
                <xsl:sequence select="$E/xs:complexType" />
            </xsl:when>
            <xsl:when test="$E/@type">
                <xsl:attribute name="type"><xsl:value-of select="$E/@type" /></xsl:attribute>
                <xsl:attribute name="idref"><xsl:value-of select="$E/@idref" /></xsl:attribute>
            </xsl:when>
            <xsl:when test="$E/@substitutionGroup">
                <xsl:sequence select="s:type($E/key('ids', @substitutionGroupRef))" />
            </xsl:when>
            <xsl:otherwise>
                <xs:complexType mixed="true">
                    <xs:sequence>
                        <xs:any minOccurs="0" maxOccurs="unbounded"/>
                    </xs:sequence>
                    <xs:anyAttribute/>
                </xs:complexType>
            </xsl:otherwise>
        </xsl:choose>
        
    </xsl:function>
    
    
    <xsl:template match="xs:attribute[not(@ref)][not(@type)][not(xs:simpleType)]" mode="elements-and-attributes-with-no-type">
        <xsl:copy>
            <xsl:copy-of select="@*" />
            <xsl:variable name="prefix" select="prefix-from-QName(node-name(.))" />
            <xsl:attribute name="type"><xsl:value-of select="concat($prefix, if ($prefix eq '') then 'anySimpleType' else ':anySimpleType')" /></xsl:attribute>
            <xsl:attribute name="idref"><xsl:value-of select="//xs:simpleType[@name eq 'anySimpleType']/@_id" /></xsl:attribute>
            <xsl:sequence select="node()" />
        </xsl:copy>
    </xsl:template>
    
    
    <xsl:template match="node()" mode="elements-and-attributes-with-no-type">
        <xsl:copy>
            <xsl:copy-of select="@*"/>
            <xsl:apply-templates mode="elements-and-attributes-with-no-type" />
        </xsl:copy>
    </xsl:template>

</xsl:stylesheet>
