<?xml version="1.0" encoding="UTF-8"?>
<xsl:stylesheet xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
                xmlns:sic="http://www.schema-library.org/schema-in-context"
                xmlns:xs="http://www.w3.org/2001/XMLSchema"
                xmlns:h="http://www.schema-library.org/input-schema-and-its-dependency-hierarchy"
                exclude-result-prefixes="#all"
                version="2.0">
    
    <!-- 
        
        Tool that Converts XML Schemas to Russian Doll
        Author: Roger L. Costello
        Version 1.0
        January 16, 2012
        
    -->
    
    <!-- 
        schema-in-context :: xs:schema -> Schema-in-Context
    -->
    
    <xsl:function name="sic:schema-in-context">
        <xsl:param name="schema" as="element()"/>
        
        <xsl:sequence select="$schema" />
        
    </xsl:function>
       
    <!-- 
        schema-in-context :: xs:schema -> Schema-in-Context -> Schema-in-Context
    -->
    
    <xsl:function name="sic:schema-in-context">
        <xsl:param name="schema" as="element()" /> 
        <xsl:param name="schema-in-context" />
        
        <xsl:sequence select="$schema" />
        <xsl:sequence select="$schema-in-context" />
        
    </xsl:function>
    
    <!-- 
        targetNamespace :: Schema-in-Context -> Maybe String
    -->
    
    <xsl:function name="sic:targetNamespace">
        <xsl:param name="schema-in-context" /> 
        
        <xsl:choose>
            <xsl:when test="exists($schema-in-context[1]/@targetNamespace)">
                <xsl:value-of select="string($schema-in-context[1]/@targetNamespace)" />
            </xsl:when>
            <xsl:otherwise/>
        </xsl:choose>
        
    </xsl:function>
    
    <!-- 
        (imports) :: Schema-in-Context -> Schema-in-Context -> Bool
    -->
    
    <xsl:function name="sic:imports" as="xs:boolean">
        <xsl:param name="parent-schema-in-context" /> 
        <xsl:param name="child-schema-in-context" />
        
        <xsl:sequence select="exists($parent-schema-in-context[1]/xs:import[@idref eq $child-schema-in-context[1]/@_id])" />
        
    </xsl:function>
    
    <!-- 
        schema-without-context :: Schema-in-Context -> xs:schema
    -->
    
    <xsl:function name="sic:schema-without-context">
        <xsl:param name="schema-in-context" /> 
        
        <xsl:sequence select="$schema-in-context[1]" />
        
    </xsl:function>
    
    <!-- 
        context :: Schema-in-Context -> Maybe Schema-in-Context
    -->
    
    <xsl:function name="sic:context">
        <xsl:param name="schema-in-context" /> 
        
        <xsl:sequence select="$schema-in-context[position() gt 1]" />
        
    </xsl:function>
    
    <!-- 
        non-import-context :: Schema-in-Context -> Maybe Schema-in-Context
        
        non-import-context s = case context s of
                                  Just s’ -> case s’ `imports` s of
                                                True -> Nothing
                                                False -> Just s’
                                  Nothing -> Nothing     
    -->
    
    <xsl:function name="sic:non-import-context">
        <xsl:param name="schema-in-context" /> 
        
        <xsl:variable name="context" select="sic:context($schema-in-context)" />
            
        <xsl:choose>
            <xsl:when test="exists($context)">
                <xsl:choose>
                    <xsl:when test="sic:imports($context, $schema-in-context)"/>
                    <xsl:otherwise>
                        <xsl:sequence select="$context" />
                    </xsl:otherwise>
                </xsl:choose>               
            </xsl:when>
            <xsl:otherwise/>
        </xsl:choose>
        
    </xsl:function>
    
    <!-- 
        namespace :: Schema-in-Context -> Maybe Namespace 
        
        namespace s = case targetNamespace s of
                         Just n -> n
                         Nothing -> let c = non-import-context s in
                                    case c of
                                       Just s’ -> namespace s’ 
                                       Nothing -> Nothing
        
    -->
    
    <xsl:function name="sic:namespace">
        <xsl:param name="schema-in-context" /> 
        
        <xsl:variable name="tns" select="sic:targetNamespace($schema-in-context)" />
        <xsl:variable name="nic" select="sic:non-import-context($schema-in-context)" />
        
        <xsl:choose>
            <xsl:when test="exists($tns)">
                <xsl:sequence select="$tns" />
            </xsl:when>
            <xsl:when test="exists($nic)">
                <xsl:sequence select="sic:namespace($nic)" />
            </xsl:when>
            <xsl:when test="empty($tns) and empty($nic)">
                <xsl:sequence select="''" />
            </xsl:when>
            <xsl:otherwise>
                <xsl:message terminate="yes">
                    Error in sic:namespace. Unable to determine the namespace of the schema. Line 155
                </xsl:message>
            </xsl:otherwise>
        </xsl:choose>
        
    </xsl:function>
    
    <!-- 
        global-items :: Schema-in-Context -> [xs:*]
        
    -->
    
    <xsl:function name="sic:global-items">
        <xsl:param name="schema-in-context" /> 
        
        <xsl:sequence select="$schema-in-context[1]/xs:*" />
        
    </xsl:function>
    
    <!-- 
        name :: Schema-in-Context -> String
        
    -->
    
    <xsl:function name="sic:name" as="xs:string">
        <xsl:param name="schema-in-context" /> 
        
        <xsl:value-of select="string($schema-in-context[1]/@_id)" />
        
    </xsl:function>
    
    <!-- 
        xsd-simpleTypes-schema :: Schema-in-Context -> xs:schema
        
    -->
     
    <xsl:function name="sic:xsd-simpleTypes-schema" as="element(xs:schema)">
        
        <xsl:sequence select="$schemas/schemas/xs:schema[ends-with(@xml:base, 'schema-data-types.xsd')]" />
        
    </xsl:function>
    
    <!-- 
        non-imported-schemas :: Schema-in-Context -> [xs:schema]
            
    --> 
    
    <xsl:function name="sic:non-imported-schemas" as="element(xs:schema)*">
        <xsl:param name="schema-in-context" as="element(xs:schema)+" /> 
        
        <xsl:for-each select="$schemas/schemas/xs:schema[@_id = $schema-in-context[1]/xs:include/@idref] |
                              $schemas/schemas/xs:schema[@_id = $schema-in-context[1]/xs:redefine/@idref]">
            
            <xsl:sequence select="." />
            
        </xsl:for-each>
        
    </xsl:function>
    
    <!-- 
        in-scope-schemas :: Schema-in-Context -> [xs:schema]
        
        in-scope-schemas s = [s, in-scope-schemas (non-imported-schemas s), in-scope-schemas (non-import-context s)]
        
        The in-scope schemas for schema s are those schemas that collectively contain s’s components. Schemas s’s in-scope schemas are:
        
        1. if s has no included schemas or non-importing context schema then s
        2. else s : in-scope-schemas ((included schemas | non-importing schema) - s)
        
    -->
    
    <xsl:function name="sic:in-scope-schemas" as="element(xs:schema)+">
        <xsl:param name="schema-in-context" as="element(xs:schema)+" /> 
        
        <xsl:variable name="found" select="sic:schema-without-context($schema-in-context)" />
        <xsl:sequence select="sic:in-scope-schemas($schema-in-context, $found)" />
        
    </xsl:function> 
    
    <xsl:function name="sic:in-scope-schemas" as="element(xs:schema)+">
        <xsl:param name="schema-in-context" />
        <xsl:param name="found" as="element(xs:schema)*"/>
        
        <xsl:variable name="item-to-process" as="element(item)">
            <item>
                <schema-to-process>
                    <xsl:sequence select="sic:schema-without-context($schema-in-context)" />
                </schema-to-process>
                <schema-in-context>
                    <xsl:sequence select="$schema-in-context" />
                </schema-in-context>
            </item>
        </xsl:variable>
        
        <xsl:apply-templates select="$process-graph">
            <xsl:with-param name="items-to-process" select="$item-to-process" />
            <xsl:with-param name="items-already-processed-or-in-queue-for-processing" select="$item-to-process" />
            <xsl:with-param name="function-to-process-current-item" select="$function-to-process-current-item" />
            <xsl:with-param name="function-to-determine-new-items-to-process" select="$function-to-determine-new-items-to-process" />
        </xsl:apply-templates>
    </xsl:function>
    
    
    <xsl:variable name="function-to-determine-new-items-to-process" select="document('')//sic:function-to-determine-new-items-to-process[1]" />
    <xsl:variable name="function-to-process-current-item" select="document('')//sic:function-to-process-current-item[1]" />
    
    <sic:function-to-determine-new-items-to-process />
    <sic:function-to-process-current-item />
    
    <xsl:template match="sic:function-to-determine-new-items-to-process">
        <xsl:param name="item" />
        <xsl:param name="items-already-processed-or-in-queue-for-processing" />
        
        <xsl:variable name="schema-to-process" select="$item/schema-to-process/xs:schema" as="element(xs:schema)"/>
        <xsl:variable name="schema-in-context" select="$item/schema-in-context/*" />
        <xsl:variable name="context" select="sic:context($schema-in-context)"/>
        
        <xsl:variable name="non-imported-schemas" select="sic:non-imported-schemas($schema-in-context)" as="element(xs:schema)*" />
        
        <xsl:variable name="nic" select="sic:non-import-context($schema-in-context)" />
        <xsl:variable name="nic-schema" select="sic:schema-without-context($nic)" as="element(xs:schema)?" />
        
        <xsl:for-each select="$non-imported-schemas[not(@_id = $items-already-processed-or-in-queue-for-processing/schema-to-process/xs:schema/@_id)]">
            <item>
                <schema-to-process>
                    <xsl:sequence select="." />
                </schema-to-process>
                <schema-in-context>
                    <xsl:sequence select="sic:schema-in-context(., $schema-in-context)" />
                </schema-in-context>
            </item>
        </xsl:for-each>
        <xsl:for-each select="$nic-schema[not(@_id = $items-already-processed-or-in-queue-for-processing/schema-to-process/xs:schema/@_id or @_id = $non-imported-schemas/@_id)]">
            <item>
                <schema-to-process>
                    <xsl:sequence select="." />
                </schema-to-process>
                <schema-in-context>
                    <xsl:sequence select="$context" />
                </schema-in-context>
            </item>
        </xsl:for-each>
        
    </xsl:template>
    
    <xsl:template match="sic:function-to-process-current-item">
        <xsl:param name="item" />
        
        <xsl:sequence select="$item/schema-to-process/xs:schema" />
        
    </xsl:template> 
    
    <xsl:function name="sic:in-scope-schemas-with-namespace" as="element(xs:schema)*">
        <xsl:param name="schema-in-context" />
        <xsl:param name="ns" as="xs:anyURI" />
        
        
        <xsl:variable name="namespace" select="sic:namespace($schema-in-context)" />
        <xsl:variable name="schema" select="sic:schema-without-context($schema-in-context)" />
        
        <xsl:variable name="import" select="sic:global-items($schema-in-context)/self::xs:import[@namespace eq $ns][@schemaLocation]" />
        
        <xsl:choose>
            <xsl:when test="$ns eq $namespace">
                <xsl:sequence select="sic:in-scope-schemas($schema-in-context)" />
            </xsl:when>
            <!-- This xsl:when was added on November 10, 2012.
                 It was added to deal with the case of a schema that *imports*
                 a schema with no namespace, i.e., import namespace="" schemaLocation="..."  -->
            <xsl:when test="exists($import)">
                <xsl:variable name="imported-schema" select="$schemas/schemas/xs:schema[@_id eq $import/@idref]" />
                <xsl:variable name="sic" select="sic:schema-in-context($imported-schema, $schema-in-context)" />
                <xsl:sequence select="sic:in-scope-schemas($sic)" />
            </xsl:when>
            <xsl:when test="$ns eq ''">  <!-- THIS IS A HACK. THE USER OF THIS FUNCTION SHOULD CONVERT ns TO namespace. SEE rewrite.xsd FOR THE PROBLEM. -->
                <xsl:sequence select="sic:in-scope-schemas($schema-in-context)" />
            </xsl:when>
            <xsl:otherwise>
                <xsl:message terminate="yes">
                    <xsl:text>[ ERROR IN sic:in-scope-schemas-with-namespace ]</xsl:text>
                    <ns><xsl:value-of select="$ns" /></ns>
                    <namespace><xsl:value-of select="$namespace" /></namespace>
                </xsl:message>
                <!--xsl:variable name="import" select="sic:global-items($schema-in-context)/self::xs:import[@namespace eq $ns][@schemaLocation]" />
                <xsl:if test="exists($import)">
                    <xsl:variable name="imported-schema" select="$schemas/schemas/xs:schema[@_id eq $import/@idref]" />
                    <xsl:variable name="sic" select="sic:schema-in-context($imported-schema, $schema-in-context)" />
                    <xsl:sequence select="sic:in-scope-schemas($sic)" />
                </xsl:if-->
            </xsl:otherwise>
        </xsl:choose>
        
    </xsl:function> 
    
</xsl:stylesheet>